@@ -1,5 +1,7 @@
 from .request import Request
 from .response import Response
+from .websocket_server import WebsocketServer
+from .websocket_client import WebsocketClient
 
 from typing import Callable, Type
 import socket
@@ -7,6 +9,8 @@ from select import select
 from pathlib import Path
 import os
 from datetime import datetime
+from hashlib import sha1
+from base64 import b64encode
 
 class Tea:
 
@@ -17,8 +21,9 @@ class Tea:
         self.__s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
         self.__s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
         
-        self.__rules  = []
-        self.__static = []
+        self.__rules     = []
+        self.__static    = []
+        self.__ws_rule   = {}
         
         
     def parse_path(self, path: str) -> dict:
@@ -81,7 +86,42 @@ class Tea:
         Add new rule on path for all valid methods. Including GET, POST, PUT, DELETE, PATCH, OPTIONS etc.
         """
         self.__rules.append({ "method": "ALL", "path": self.parse_path(path), "callback": callback })
+            
+    
+    def websocket(self, path: str) -> Type[WebsocketServer]:
+        ws_server      = WebsocketServer()
+        self.__ws_rule = { "path": self.parse_path(path), "ws_server": ws_server }
+        return ws_server
+    
+    
+    def __do_ws_handshake(self, req: Type[Request], res: Type[Response], conn: Type[socket.socket]) -> bool:
+        # returns success or fail state
+        # check if its a valid ws handshake request
+        is_valid = req.method == "GET" and\
+            float(req.http_version.split("/")[1]) >= 1.1 and\
+            req.has_header("upgrade") and\
+            req.get_header("upgrade").lower() == "websocket" and\
+            req.has_header("connection") and\
+            req.get_header("connection").lower() == "upgrade" and\
+            req.has_header("Sec-WebSocket-Key")
         
+        if not is_valid:
+            res.send("400 Bad Request", status_code=400)
+            conn.sendall(res.get_res_as_text().encode())
+            self.__close_conn(conn)
+            return 0
+        
+        # do the handshake if valid
+        ws_client_key = req.get_header("Sec-WebSocket-Key")
+        hashed = sha1((ws_client_key + "258EAFA5-E914-47DA-95CA-C5AB0DC85B11").encode())
+        ws_accept_key = b64encode(hashed.digest()).decode()
+        res.send(headers={ "Upgrade": "Websocket", "Connection": "Upgrade", "Sec-WebSocket-Accept": ws_accept_key }, status_code=101)
+        self.__ws_sockets.append(conn)
+        self.__ws_rule["ws_server"].add_client(WebsocketClient(conn))
+        conn.sendall(res.get_res_as_text().encode())
+        # not closing the websocket connection
+        return 1
+    
         
     def __handle_req(self, req: Type[Request], conn: Type[socket.socket]) -> None:
         req = Request(req)
@@ -91,6 +131,16 @@ class Tea:
         if self.__mode == "development":
             print(f"[{datetime.now().strftime('%X - %x')}] > {req.method} http://{self.__host}:{self.__port}{req.url.href}")
         
+        # check if its websocket handshake request
+        if self.__ws_rule:
+            if req.url.pathname == self.__ws_rule["path"]["pathname"]:
+                success = self.__do_ws_handshake(req, res, conn)
+                ws_client = WebsocketClient(socket)
+                if success:
+                    event = self.__ws_rule["ws_server"].Event(ws_client, 1)
+                    self.__ws_rule["ws_server"].onopen(event)
+                return
+        
         # check if path is served as a static file
         static_file = list(filter(lambda f: f"{str(f['file'].expanduser()).replace(f['folder_path'], '' if len(f['path']) == 1 else f['path'], 1)}" == req.url.pathname, self.__static))
         if len(static_file) > 0:
@@ -131,8 +181,7 @@ class Tea:
                 if is_matched_real and len(matched_rules) > 0:
                     # most recently added rule takes precedence
                     matched_rules.reverse()
-                    res.body = "405 Method Not Allowed"
-                    res.status_code = 405
+                    res.send("405 Method Not Allowed", status_code=405)
                     
                     # rule with specific method takes precedence against ALL
                     is_matched = False
@@ -159,7 +208,13 @@ class Tea:
                             
             
         conn.sendall(res.get_res_as_text().encode())
-     
+        self.__close_conn(conn)
+    
+    
+    def __close_conn(self, conn: Type[socket.socket]):
+        self.__in_sockets.remove(conn)
+        conn.close()
+        
      
     def listen(self, host: str="127.0.0.1", port: int=5500, mode: str="development") -> None:
         """
@@ -178,25 +233,38 @@ class Tea:
         self.__in_sockets  = [self.__s]
         out_sockets        = [] # actually, we're not gonna use this
         x_sockets          = [] # this too
+        self.__ws_sockets  = []
 
         while 1:
             try:
                 readable, _, _ = select(self.__in_sockets, out_sockets, x_sockets)
-                for client in readable:
+                for readable in readable:
                     # continue if connection is already closed
-                    if client.fileno() == -1: continue
+                    if readable.fileno() == -1: continue
                     
                     # handle new connection
-                    if client is self.__s:
+                    if readable is self.__s:
                         conn, addr = self.__s.accept()
                         self.__in_sockets.append(conn)
+                    
+                    # handle websocket message
+                    # thanks to https://github.com/AlexanderEllis/websocket-from-scratch
+                    elif readable in self.__ws_sockets:
+                        ws_client = WebsocketClient(readable)
+                        msg = ws_client.read(self.max_buffer_size)
+                        if msg:
+                            event = self.__ws_rule["ws_server"].Event(ws_client, 1, msg)
+                            self.__ws_rule["ws_server"].onmessage(event)
+                        else:
+                            self.__close_conn(readable)
+                            self.__ws_sockets.remove(readable)
+                            event = self.__ws_rule["ws_server"].Event(ws_client, 3)
+                            self.__ws_rule["ws_server"].onclose(event)
                         
-                    # handle the request
-                    else:                        
-                        req = client.recv(self.max_buffer_size)
-                        self.__handle_req(req.decode(), client)
-                        self.__in_sockets.remove(client)
-                        client.close()
+                    # handle regular http request
+                    else:
+                        req = readable.recv(self.max_buffer_size)
+                        self.__handle_req(req.decode(), readable)
 
             except KeyboardInterrupt:
                 self.__s.close()
