@@ -2,11 +2,11 @@ from urllib.parse import urlparse
 
 class URL:
     
-    def __init__(self, url):
+    def __init__(self, url: str):
         self.parse_url(url)
 
 
-    def parse_url(self, url):
+    def parse_url(self, url: str) -> None:
         url = urlparse(url)
         # customized to make it more like javascript URL standard
         # see https://developer.mozilla.org/en-US/docs/Web/API/URL for reference
